!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   rg_generate_mod
!> @brief   generate RPATH input files
!! @authors Norio Takase (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../../config.h"
#endif

module rg_generate_mod

  use rg_option_mod
  use rg_option_str_mod
  use fileio_trj_mod
  use output_str_mod
  use input_str_mod
  use trajectory_str_mod
  use fileio_grocrd_mod
  use fileio_prmtop_mod
  use fileio_pdb_mod
  use fileio_rst_mod
  use fileio_mod
  use string_mod
  use messages_mod
  use constants_mod

  implicit none
  private

  ! structures
  type :: s_cv
    real(wp), allocatable :: ts(:)
    real(wp), allocatable :: cv(:,:)
  end type s_cv

  type :: s_idxrep
    integer               :: idx
    integer               :: rep
  end type s_idxrep

  ! subroutine
  public  :: generate
  private :: open_cv
  private :: compute_path
  private :: reparametrize_path
  private :: output_snapshot
  private :: output_snapshot_pdb
  private :: output_snapshot_rst
  private :: output_control
  private :: check_reference
  private :: get_distance
  private :: get_numbered_filename
  private :: sort_idxrep
  private :: cumsum

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    generate
  !> @brief        generate molecule files
  !! @authors      NT
  !! @param[in]    input    : input information
  !! @param[in]    output   : output information
  !! @param[in]    option   : option information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine generate(input, output, option)

    ! formal arguments
    type(s_input),           intent(in)    :: input
    type(s_output),          intent(in)    :: output
    type(s_option),          intent(in)    :: option

    ! local variables
    type(s_cv)               :: cv
    type(s_pdb)              :: ref
    real(wp),   allocatable  :: path(:,:), path_equi(:,:)


    call open_cv(input%cvfile, cv)


    call compute_path(option%nreplica, cv, path)


    call reparametrize_path(path, path_equi)


    call check_reference(input%reffile,    &
                         input%prmtopfile, &
                         input%grocrdfile, &
                         ref)


    call output_snapshot(ref, cv, path_equi, &
                         input%dcdfile,    &
                         input%dcdvelfile, &
                         output%pdbfile,   &
                         output%rstfile,   &
                         option)


    call output_control (path_equi)


    call dealloc_pdb_all(ref)
    deallocate(path, path_equi)

    return

  end subroutine generate

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine open_cv(cvfile, cv)

    ! formal arguments
    character(*),            intent(in)    :: cvfile
    type(s_cv),              intent(inout) :: cv

    ! local variables
    integer                  :: file, ndim, nstep, i, j
    character(1000)          :: line


    call open_file(file, cvfile, IOFileInput)

    nstep = 1

    read(file,'(a)') line
    ndim = split_num(line)
    ndim = ndim - 1

    do while(.true.)
      read(file,'(a)',end=10,err=10) line
      nstep = nstep + 1
    end do

10  write(MsgOut,'(a,i8)') 'Open_Cv> # of dimension : ', ndim
    write(MsgOut,'(a,i8)') '         # of steps     : ', nstep
    write(MsgOut,'(a)')    ' '

    allocate(cv%ts(nstep), cv%cv(ndim,nstep))

    rewind(file)

    do i = 1, nstep
      read(file,*) cv%ts(i),(cv%cv(j,i),j=1,ndim)
    end do

    call close_file(file)

    return

  end subroutine open_cv

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_path(nreplica, cv, path)

    ! formal arguments
    integer,                 intent(in)    :: nreplica
    type(s_cv),              intent(inout) :: cv
    real(wp),  allocatable,  intent(inout) :: path(:,:)

    ! local variables
    integer                  :: ndim, nstep, ireplica, i, count
    real(wp),  allocatable   :: accum(:)


    ndim  = size(cv%cv(:,1))
    nstep = size(cv%cv(1,:))

    allocate(path(ndim, nreplica))
    allocate(accum(ndim))

    path(:,1)        = cv%cv(:,1)
    path(:,nreplica) = cv%cv(:,nstep)

    cv%ts(:) = (cv%ts(:) / maxval(cv%ts(:))) * (nreplica - 1)

    do ireplica = 1, nreplica-2
      accum(:) = 0
      count = 0
      do i = 1, nstep
        if (ireplica-0.5_wp <= cv%ts(i) .and. &
            cv%ts(i) < (ireplica + 0.5_wp)) then
          accum(:) = accum(:) + cv%cv(:,i)
          count = count + 1
        end if
      end do
      path(:,ireplica+1) = accum / real(count,wp)
    end do

    deallocate(accum)

    return

  end subroutine compute_path

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine reparametrize_path(path, path_equi)

    ! formal arguments
    real(wp),                intent(in)    :: path(:,:)
    real(wp), allocatable,   intent(inout) :: path_equi(:,:)

    ! local variables
    real(wp)                 :: d
    integer                  :: ndim, nreplica, i, q

    real(wp), allocatable    :: dist(:), dist_cumsum(:), dist_equi(:)


    ndim     = size(path(:,1))
    nreplica = size(path(1,:))

    allocate(path_equi(ndim, nreplica))
    allocate(dist(nreplica), dist_cumsum(nreplica), dist_equi(nreplica))

    dist(1) = 0.0_wp
    do i = 2, nreplica
      d = get_distance(path(:,i), path(:,i-1))
      dist(i) = d
    end do

    call cumsum(dist, dist_cumsum)

    do i = 1, nreplica
      d = dist_cumsum(nreplica)*real(i-1,wp)/real(nreplica-1)
      dist_equi(i) = d
    end do

    path_equi(:,1) = path(:,1)
    do i = 2, nreplica - 1
      q = 2
      do while (dist_cumsum(q) < dist_equi(i))
        q = q + 1
      end do
      path_equi(:,i) = path(:,q-1) + (path(:,q)-path(:,q-1)) * &
                                     (dist_equi(i) - dist_cumsum(q-1))/ &
                                     (dist_cumsum(q) - dist_cumsum(q-1))
      
    end do
    path_equi(:,nreplica) = path(:,nreplica)

    deallocate(dist, dist_cumsum, dist_equi)

    return

  end subroutine reparametrize_path

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_snapshot(ref, cv, path, &
                             dcdfile, dcdvelfile, pdbfile, rstfile, option)

    ! formal arguments
    type(s_pdb),             intent(inout) :: ref
    type(s_cv),              intent(in)    :: cv
    real(wp),                intent(in)    :: path(:,:)
    character(*),            intent(in)    :: dcdfile
    character(*),            intent(in)    :: dcdvelfile
    character(*),            intent(in)    :: pdbfile
    character(*),            intent(in)    :: rstfile
    type(s_option),          intent(in)    :: option
    
    ! local variables
    real(wp)                 :: d, dmin
    integer                  :: nstep, nreplica
    integer                  :: i, j, istep, ireplica, idx, prv_idx
    type(s_trj_file)         :: dcd, dvl

    type(s_trajectory)       :: trjcrd, trjvel
    type(s_idxrep), allocatable :: idxrep(:)


    ! open trajectory
    !
    dcd%unit_no = 0
    dvl%unit_no = 0

    if (dcdfile /= '') then
      call open_trj(dcd, dcdfile, TrjFormatDCD, TrjTypeCoorBox, IOFileInput)
      call alloc_trajectory(trjcrd, ref%num_atoms)
    else
      call error_msg('Output_Snapshot> Dcd file must be given.')
    end if

    if (dcdvelfile /= '') then
      call open_trj(dvl, dcdvelfile, TrjFormatDCD, TrjTypeCoorBox, IOFileInput)
      call alloc_trajectory(trjvel, ref%num_atoms)
    end if


    ! select trajectory step for replica
    !
    nreplica = size(path(1,:))
    nstep    = size(cv%cv(1,:))-2  ! first and last step is not used

    allocate(idxrep(nreplica))

    do ireplica = 1, nreplica

      dmin = 100000.0_wp
      idx  = -1
      do istep = 1, nstep
        d = get_distance(path(:,ireplica),cv%cv(:,istep+1))
        if (d < dmin) then
          dmin = d
          idx  = istep
        end if
      end do

      write(MsgOut,'(a,i5,a,i5,a)') &
           'Output_Snapshot> ', idx,      ' th snapshot is used for ', &
                                ireplica, ' th replica'

      idxrep(ireplica)%idx = idx
      idxrep(ireplica)%rep = ireplica

    end do
    write(MsgOut,'(a)') ''

    call sort_idxrep(idxrep, 1, nreplica)


    ! read trajectory and write restart/pdb file
    !
    write(MsgOut,'(a)') 'Output_SnapShot> write restart/pdb file'

    prv_idx = 1

    do i = 1, nreplica

      idx      = idxrep(i)%idx
      ireplica = idxrep(i)%rep

      write(MsgOut,'(a,i6,a,i6)') &
           '   snapshot : ', idx, '  replica : ', ireplica

      if (dcd%unit_no /= 0) then
        do j = prv_idx, idx
          call read_trj(dcd, trjcrd)
        end do
      end if

      if (dvl%unit_no /= 0) then
        do j = prv_idx, idx
          call read_trj(dvl, trjvel)
        end do
      end if

      if (pdbfile /= '') &
        call output_snapshot_pdb &
            (ref, trjcrd,         ireplica, idx, prv_idx, pdbfile, option)

      if (rstfile /= '') &
        call output_snapshot_rst &
            (ref, trjcrd, trjvel, ireplica, idx, prv_idx, rstfile, option)

      prv_idx = idx+1

    end do

    deallocate(idxrep)

    write(MsgOut,'(a)') ''


    ! close trjectory
    !
    call dealloc_trajectory(trjcrd)
    call dealloc_trajectory(trjvel)

    if (dcdfile /= '') &
      call close_trj(dcd)

    if (dcdvelfile /= '') &
      call close_trj(dvl)

    return

  end subroutine output_snapshot

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_snapshot_pdb(ref, trjcrd, &
                                 irep, idx, idx0, pdbfile, option)

    ! formal arguments
    type(s_pdb),             intent(inout) :: ref
    type(s_trajectory),      intent(in)    :: trjcrd
    integer,                 intent(in)    :: irep
    integer,                 intent(in)    :: idx
    integer,                 intent(in)    :: idx0
    character(*),            intent(in)    :: pdbfile
    type(s_option),          intent(in)    :: option
    

    ref%atom_coord(1:3,1:ref%num_atoms) = trjcrd%coord(1:3,1:ref%num_atoms)

    call output_pdb(get_numbered_filename(pdbfile, irep), ref)

    return

  end subroutine output_snapshot_pdb

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_snapshot_rst(ref, trjcrd, trjvel, &
                                 irep, idx, idx0, rstfile,option)

    ! formal arguments
    type(s_pdb),             intent(in)    :: ref
    type(s_trajectory),      intent(in)    :: trjcrd
    type(s_trajectory),      intent(in)    :: trjvel
    integer,                 intent(in)    :: irep
    integer,                 intent(in)    :: idx
    integer,                 intent(in)    :: idx0
    character(*),            intent(in)    :: rstfile
    type(s_option),          intent(in)    :: option
    
    ! local variables
    type(s_rst)              :: rst


    call alloc_rst(rst, RestartAtom, ref%num_atoms)

    if (allocated(trjvel%coord)) then

      ! md restart file

      rst%num_atoms                     = ref%num_atoms
      rst%rstfile_type                  = RstfileTypeMd
      rst%iseed                         = option%md_iseed
      rst%box_size_x                    = trjcrd%pbc_box(1,1)
      rst%box_size_y                    = trjcrd%pbc_box(2,2)
      rst%box_size_z                    = trjcrd%pbc_box(3,3)
      rst%thermostat_momentum           = option%md_thermo_moment
      rst%barostat_momentum(1:3)        = option%md_baro_moment
      rst%coord   (1:3,1:ref%num_atoms) = trjcrd%coord(1:3,1:ref%num_atoms)
      rst%velocity(1:3,1:ref%num_atoms) = trjvel%coord(1:3,1:ref%num_atoms)

      call output_rst(get_numbered_filename(rstfile, irep), rst)

    else

      ! min restart file

      rst%num_atoms                     = ref%num_atoms
      rst%rstfile_type                  = RstfileTypeMin
      rst%energy                        = option%min_energy
      rst%delta_r                       = option%min_delta_r
      rst%box_size_x                    = trjcrd%pbc_box(1,1)
      rst%box_size_y                    = trjcrd%pbc_box(2,2)
      rst%box_size_z                    = trjcrd%pbc_box(3,3)
      rst%coord   (1:3,1:ref%num_atoms) = trjcrd%coord(1:3,1:ref%num_atoms)

      call output_rst(get_numbered_filename(rstfile, irep), rst)

    end if

    call dealloc_rst_all(rst)


    return

  end subroutine output_snapshot_rst

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_control(path)

    ! formal arguments
    real(wp),                intent(in)    :: path(:,:)

    ! local variables
    integer                  :: nreplica, ndim, i, idim
    character(100)           :: fmt


    ndim     = size(path(:,1))
    nreplica = size(path(1,:))

    write(fmt,'(a,i0,a)') '(a,i0,a,',nreplica,'(f0.6,1x))'

    write(MsgOut,'(a)') &
         ''
    write(MsgOut,'(a)') &
         'copy the follwing for RPATH simulation'
    write(MsgOut,'(a)') &
         '--------------------------------------'
    write(MsgOut,'(a)') &
         '[RPATH]'
    write(MsgOut,'(a,i0)') &
         'nreplica      = ', nreplica
    write(MsgOut,'(a)') &
         'rpath_period  = 0'
    write(MsgOut,'(a,10(i0,1x))') &
         'rest_function = ', (i,i=1,ndim)
    write(MsgOut,'(a)') &
         ''
    write(MsgOut,'(a)') &
         '[RESTRAINTS]'
    write(MsgOut,'(a,i0)') &
         'nfunctions     = ', ndim
    write(MsgOut,'(a)') &
         ''
    do idim = 1, ndim
      write(MsgOut,'(a,i0,a)') &
         'function', idim, '     = DIHED'
      write(MsgOut,fmt) &
         'constant', idim, '     = ', (1.0_wp,i=1,nreplica)
      write(MsgOut,fmt) &
         'reference', idim, '    = ', (path(idim,i),i=1,nreplica)
      write(MsgOut,'(a,i0,a)') &
         'select_index', idim, ' = 1 2 3 4'
      write(MsgOut,'(a)') &
         ''
    end do
    write(MsgOut,'(a)') &
         ''

    return

  end subroutine output_control

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine check_reference(reffile, prmtopfile, grocrdfile, ref)

    ! formal arguments
    character(*),            intent(in)    :: reffile
    character(*),            intent(in)    :: prmtopfile
    character(*),            intent(in)    :: grocrdfile
    type(s_pdb),             intent(inout) :: ref

    ! local variables
    type(s_prmtop)           :: prmtop
    type(s_grocrd)           :: grocrd
    integer                  :: i


    if (reffile == '' .and. prmtopfile == '' .and. grocrdfile == '') &
      call error_msg('Check_Reference> reference file is not given.')

    ! with pdb
    if (reffile /= '') then

      call input_pdb(reffile, ref)

    end if

    ! with prmtop
    if (prmtopfile /= '') then

      call input_prmtop(prmtopfile, prmtop)

      call init_pdb(ref)
      ref%hetatm_rec = .true.
      ref%cryst_rec  = .true.
      ref%atom_col7  = .true.
      ref%res_col5   = .true.
      ref%num_atoms  = prmtop%num_atoms

      call alloc_pdb(ref, PdbAtom, ref%num_atoms)

      ! atom_name
      ref%atom_name(:) = prmtop%atom_name(:)

      ! atom no
      do i = 1, ref%num_atoms
        ref%atom_no(i) = i
      end do

      ! residue name
      ! residue no
      do i = 1, size(prmtop%res_label)-1
        ref%residue_name(prmtop%res_point(i):prmtop%res_point(i+1)-1) = &
             prmtop%res_label(i)
        ref%residue_no  (prmtop%res_point(i):prmtop%res_point(i+1)-1) = &
             i
      end do
      i = size(prmtop%res_label)
      ref%residue_name(prmtop%res_point(i):) = &
           prmtop%res_label(i)
      ref%residue_no  (prmtop%res_point(i):) = &
           i

    end if

    ! with grocrd
    if (grocrdfile /= '') then

      call input_grocrd(grocrdfile, grocrd)

      call init_pdb(ref)
      ref%hetatm_rec = .true.
      ref%cryst_rec  = .true.
      ref%atom_col7  = .true.
      ref%res_col5   = .true.
      ref%num_atoms  = grocrd%num_atoms

      call alloc_pdb(ref, PdbAtom, ref%num_atoms)

      ! atom_no
      ref%atom_no(:) = grocrd%atom_no(:)

      ! atom_name
      ref%atom_name(:) = grocrd%atom_name(:)

      ! residue_no
      ref%residue_no(:) = grocrd%residue_no(:)

      ! residue_name
      ref%residue_name(:) = grocrd%residue_name(:)

    end if

    return

  end subroutine check_reference

  !======1=========2=========3=========4=========5=========6=========7=========8

  function get_distance(p1, p2)

    ! return values
    real(wp)                 :: get_distance

    ! formal arguments
    real(wp),                intent(in)    :: p1(:)
    real(wp),                intent(in)    :: p2(:)

    ! local variables
    real(wp)                 :: accum, d
    integer                  :: i
    

    accum = 0.0_wp

    do i = 1, size(p1)
      d = p2(i) - p1(i)
      accum = accum + d * d
    end do
    get_distance = sqrt(accum)

    return

  end function get_distance

  !======1=========2=========3=========4=========5=========6=========7=========8

  function get_numbered_filename(basename, num)

    ! return value
    character(1000)          :: get_numbered_filename

    ! formal arguments
    character(*),            intent(in)    :: basename
    integer,                 intent(in)    :: num

    ! local variables
    integer                  :: lidx, ridx
    character(1000)          :: lstr, rstr


    lidx = index(basename, '{', back=.true.)
    ridx = index(basename, '}', back=.true.)

    if (lidx == 0 .or. ridx == 0) then
      get_numbered_filename = basename
      return
    end if

    lstr = basename(:lidx-1)
    rstr = basename(ridx+1:)

    write(get_numbered_filename, '(a,i0,a)') trim(lstr), num, trim(rstr)

    return

  end function get_numbered_filename

  !======1=========2=========3=========4=========5=========6=========7=========8

  recursive subroutine sort_idxrep(a, start, end)

    ! formal arguments
    type(s_idxrep),          intent(inout) :: a(*)
    integer,                 intent(in)    :: start
    integer,                 intent(in)    :: end

    ! local variables
    type(s_idxrep)           :: t, x
    integer                  :: i, j


    x = a((start + end) / 2)
    i = start
    j = end

    do
      do while (a(i)%idx < x%idx)
        i = i + 1
      end do

      do while (x%idx < a(j)%idx)
        j = j - 1
      end do

      if (i >= j) exit

      t = a(i)
      a(i) = a(j)
      a(j) = t
      i = i + 1
      j = j - 1

    end do

    if (start < i - 1) &
      call sort_idxrep(a, start, i - 1)
    if (j + 1 < end) &
      call sort_idxrep(a, j + 1, end)

    return

  end subroutine sort_idxrep

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine cumsum(array_in, array_out)

    ! formal arguments
    real(wp),                intent(in)    :: array_in(:)
    real(wp),                intent(inout) :: array_out(:)

    ! local variables
    integer                  :: i
    real(wp)                 :: accum


    accum = 0.0_wp

    do i = 1, size(array_in)
      accum = accum + array_in(i)
      array_out(i) = accum
    end do

    return

  end subroutine cumsum

end module rg_generate_mod
