!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_enefunc_mod
!> @brief   define potential energy functions
!! @authors Yuji Sugita (YS), Takaharu Mori (TM), Chigusa Kobayashi (CK)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_enefunc_mod

  use at_enefunc_gromacs_mod
  use at_enefunc_amber_mod
  use at_enefunc_charmm_mod
  use at_enefunc_go_mod
  use at_energy_mod
  use at_restraints_str_mod
  use at_enefunc_str_mod
  use at_energy_str_mod
  use at_boundary_str_mod
  use molecules_str_mod
  use fileio_grotop_mod
  use fileio_prmtop_mod
  use fileio_par_mod
  use fileio_gpr_mod
  use messages_mod
  use constants_mod

  implicit none
  private

  ! subroutines
  public  :: define_enefunc
  private :: setup_enefunc_dispcorr

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    define_enefunc
  !> @brief        a driver subroutine for defining potential energy functions
  !! @authors      TM, CK
  !! @param[in]    ene_info   : ENERGY section control parameters
  !! @param[in]    boundary   : boundary conditions information
  !! @param[in]    par        : CHARMM PAR information
  !! @param[in]    gpr        : GO model parameter information
  !! @param[in]    prmtop     : AMBER parameter topology information
  !! @param[in]    grotop     : GROMACS parameter topology information
  !! @param[in]    molecule   : molecule information
  !! @param[in]    restraints : restraints information
  !! @param[out]   enefunc    : structure of enefunc
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine define_enefunc(ene_info, boundary, par, gpr, prmtop, grotop, &
                            molecule, restraints, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info 
    type(s_boundary),        intent(in)    :: boundary
    type(s_par),             intent(in)    :: par
    type(s_gpr),             intent(in)    :: gpr
    type(s_prmtop),          intent(in)    :: prmtop
    type(s_grotop),          intent(in)    :: grotop
    type(s_molecule),        intent(in)    :: molecule
    type(s_restraints),      intent(in)    :: restraints
    type(s_enefunc),         intent(inout) :: enefunc


    call init_enefunc(enefunc)


    enefunc%forcefield      = ene_info%forcefield
    enefunc%output_style    = ene_info%output_style

    enefunc%switchdist      = ene_info%switchdist
    enefunc%cutoffdist      = ene_info%cutoffdist
    enefunc%pairlistdist    = ene_info%pairlistdist
    enefunc%dielec_const    = ene_info%dielec_const
    enefunc%force_switch    = ene_info%vdw_force_switch
    enefunc%vdw_shift       = ene_info%vdw_shift
    enefunc%dispersion_corr = ene_info%dispersion_corr
    enefunc%contact_check   = ene_info%contact_check
    enefunc%minimum_contact = ene_info%minimum_contact


    if (par%num_bonds > 0) then

      call define_enefunc_charmm &
                         (ene_info, boundary, par, molecule, &
                          restraints, enefunc)

    else if (gpr%num_bonds > 0) then

      call define_enefunc_go &
                         (ene_info, gpr, molecule, restraints, enefunc)

    else if (prmtop%num_atoms > 0) then

      call define_enefunc_amber &
                          (ene_info, boundary, prmtop, molecule, &
                           restraints, enefunc)

    else if (grotop%num_atomtypes > 0) then

      call define_enefunc_gromacs &
                          (ene_info, boundary, grotop, molecule, &
                           restraints, enefunc)

    end if

    ! dispersion correction
    !
    call setup_enefunc_dispcorr(ene_info, molecule, enefunc)

    return

  end subroutine define_enefunc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_dispcorr
  !> @brief        define dispersion correction term 
  !! @authors      CK
  !! @param[in]    ene_info : ENERGY section control parameters information
  !! @param[inout] molecule : molecule information
  !! @param[inout] enefunc  : energy potential functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  subroutine setup_enefunc_dispcorr(ene_info, molecule, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)      :: ene_info
    type(s_molecule),        intent(in)      :: molecule
    type(s_enefunc), target, intent(inout)   :: enefunc

    ! local variables
    integer                  :: i, j, iatmcls, ntypes
    integer                  :: i1, i2, i3, i4
    integer                  :: natom2, nexpair
    real(wp)                 :: lj6_tot, lj6_diff, lj6_ex
    real(wp)                 :: factor, rpair
    real(wp)                 :: diff_cs, diff_cs2, diff_cs3, diff_cs4
    real(wp)                 :: cutoff , cutoff2, cutoff3, cutoff4
    real(wp)                 :: cutoff5, cutoff6, cutoff7, cutoff8
    real(wp)                 :: cutoff14
    real(wp)                 :: inv_cutoff3, inv_cutoff6, inv_cutoff12
    real(wp)                 :: switchdist , switchdist2, switchdist3
    real(wp)                 :: switchdist4, switchdist5
    real(wp)                 :: switchdist6, switchdist7, switchdist8
    real(wp)                 :: shift_a, shift_b, shift_c
    real(wp)                 :: vswitch, eswitch, vlong

    integer,         pointer :: bondlist(:,:),anglelist(:,:), ureylist(:,:)
    integer,         pointer :: dihelist(:,:),rb_dihelist(:,:)
    integer,         pointer :: imprlist(:,:)

    integer,     allocatable :: atype(:)


    if (ene_info%dispersion_corr == Disp_corr_NONE) &
      return

    bondlist    => enefunc%bond_list
    anglelist   => enefunc%angl_list
    ureylist    => enefunc%urey_list
    dihelist    => enefunc%dihe_list
    rb_dihelist => enefunc%rb_dihe_list
    imprlist    => enefunc%impr_list

    ntypes = enefunc%num_atom_cls
    allocate(atype(1:ntypes))

    atype(1:ntypes) = 0

    do i = 1, molecule%num_atoms
      iatmcls = molecule%atom_cls_no(i)
      atype(iatmcls) = atype(iatmcls)+1
    end do

    lj6_tot = 0.0_wp
    do i = 1, ntypes
      do j = 1, ntypes
        lj6_tot = lj6_tot + enefunc%nonb_lj6(i,j)*atype(i)*atype(j)
      end do
    end do

    deallocate(atype)

    cutoff      = enefunc%cutoffdist
    cutoff2     = cutoff*cutoff
    cutoff3     = cutoff2*cutoff
    inv_cutoff3 = 1.0_wp/cutoff3

    eswitch = 0.0_wp
    vswitch = 0.0_wp
    vlong   = inv_cutoff3/3.0_wp

    if (enefunc%forcefield == ForcefieldAMBER) then

      factor = 2.0_wp*PI*lj6_tot
      enefunc%dispersion_energy = -factor*vlong
      enefunc%dispersion_virial = -2.0_wp*factor*vlong

    else if (enefunc%forcefield == ForcefieldGROAMBER .or.             &
             enefunc%forcefield == ForcefieldGROMARTINI) then

      ! remove exclusion
      !
      lj6_ex = 0.0_wp
      nexpair = 0
      do i = 1, molecule%num_atoms
        iatmcls = molecule%atom_cls_no(i)
        lj6_ex  = lj6_ex + enefunc%nonb_lj6(iatmcls,iatmcls)
      end do

      ! bonds
      do i = 1, enefunc%num_bonds
        i1     = molecule%atom_cls_no(bondlist(1,i))
        i2     = molecule%atom_cls_no(bondlist(2,i))
        lj6_ex = lj6_ex + enefunc%nb14_lj6(i1,i2)
      end do

      ! angles
      do i = 1, enefunc%num_angles
        i1     = molecule%atom_cls_no(anglelist(1,i))
        i3     = molecule%atom_cls_no(anglelist(3,i))
        lj6_ex = lj6_ex + enefunc%nb14_lj6(i1,i3)
      end do

      ! Urey angles
      do i = 1, enefunc%num_ureys
        i1     = molecule%atom_cls_no(ureylist(1,i))
        i3     = molecule%atom_cls_no(ureylist(3,i))
        lj6_ex = lj6_ex + enefunc%nb14_lj6(i1,i3)
      end do

      ! dihedral
      do i = 1, enefunc%num_dihedrals
        i1     = molecule%atom_cls_no(dihelist(1,i))
        i4     = molecule%atom_cls_no(dihelist(4,i))
        lj6_ex = lj6_ex + enefunc%nb14_lj6(i1,i4)
      end do

      ! RB dihedral
      do i = 1, enefunc%num_rb_dihedrals
        i1     = molecule%atom_cls_no(rb_dihelist(1,i))
        i4     = molecule%atom_cls_no(rb_dihelist(4,i))
        lj6_ex = lj6_ex + enefunc%nb14_lj6(i1,i4)
      end do

      ! improper
      do i = 1, enefunc%num_impropers
        i1     = molecule%atom_cls_no(imprlist(1,i))
        i4     = molecule%atom_cls_no(imprlist(4,i))
        lj6_ex = lj6_ex + enefunc%nb14_lj6(i1,i4)
      end do

      nexpair = nexpair + molecule%num_atoms       &
                        + enefunc%num_bonds        &
                        + enefunc%num_angles       &
                        + enefunc%num_ureys        &
                        + enefunc%num_dihedrals    &
                        + enefunc%num_rb_dihedrals &
                        + enefunc%num_impropers

      lj6_diff = (lj6_tot - lj6_ex)

      natom2 = molecule%num_atoms*molecule%num_atoms
      rpair  = real(natom2/(natom2-nexpair),wp)
      factor       = 2.0_wp*PI*rpair*lj6_diff

      switchdist   = enefunc%switchdist
      diff_cs      = (cutoff - switchdist)

      if (diff_cs > EPS) then

        if (enefunc%vdw_shift) then

          cutoff4      = cutoff3*cutoff
          cutoff5      = cutoff4*cutoff
          cutoff6      = cutoff5*cutoff
          cutoff7      = cutoff6*cutoff
          cutoff8      = cutoff7*cutoff
          cutoff14     = cutoff7*cutoff7
          inv_cutoff6  = inv_cutoff3*inv_cutoff3
          inv_cutoff12 = inv_cutoff6*inv_cutoff6
  
          diff_cs2     = diff_cs*diff_cs
          diff_cs3     = diff_cs2*diff_cs
          diff_cs4     = diff_cs3*diff_cs
  
          switchdist2  = switchdist*switchdist
          switchdist3  = switchdist2*switchdist
          switchdist4  = switchdist3*switchdist
          switchdist5  = switchdist4*switchdist
          switchdist6  = switchdist5*switchdist
          switchdist7  = switchdist6*switchdist
          switchdist8  = switchdist7*switchdist
  
          ! LJ6
          ! 
          shift_a = -(10.0_wp*cutoff - 7.0_wp*switchdist)/(cutoff8*diff_cs2)
          shift_b =  ( 9.0_wp*cutoff - 7.0_wp*switchdist)/(cutoff8*diff_cs3)
  
          shift_c = inv_cutoff6 - 2.0_wp * shift_a * diff_cs3  &
                    - 1.5_wp * shift_b * diff_cs4
  
          eswitch = -2.0_wp * shift_a * ((1.0_wp/6.0_wp)*cutoff6               &
                                        -(3.0_wp/5.0_wp)*cutoff5*switchdist    &
                                        +(3.0_wp/4.0_wp)*cutoff4*switchdist2   &
                                        -(1.0_wp/3.0_wp)*cutoff3*switchdist3   &
                                        +(1.0_wp/6.0e1_wp)*switchdist6)        &
                    -1.5_wp * shift_b * ((1.0_wp/7.0_wp)*cutoff7               &
                                        -(2.0_wp/3.0_wp)*cutoff6*switchdist    &
                                        +(6.0_wp/5.0_wp)*cutoff5*switchdist2   &
                                        -                cutoff4*switchdist3   &
                                        +(1.0_wp/3.0_wp)*cutoff3*switchdist4   &
                                        -(1.0_wp/1.05e2_wp)*switchdist7)       &
                    -(1.0_wp/3.0_wp) * shift_c * (cutoff3-switchdist3)
  
          ! LJ12
          !
          shift_a = -(16.0_wp*cutoff - 13.0_wp*switchdist)/(cutoff14*diff_cs2)
          shift_b =  (15.0_wp*cutoff - 13.0_wp*switchdist)/(cutoff14*diff_cs3)
          shift_c = inv_cutoff12 - 2.0_wp * shift_a * diff_cs3  &
                    - 1.5_wp * shift_b * diff_cs4
  
  
          vswitch = shift_a * ( (1.0_wp/6.0_wp)*cutoff6                        &
                               -(2.0_wp/5.0_wp)*cutoff5*switchdist             &
                               +(1.0_wp/4.0_wp)*cutoff4*switchdist2            &
                               -(1.0_wp/6.0e1_wp)*switchdist6)                 &
                   +shift_b * ( (1.0_wp/7.0_wp)*cutoff7                        &
                               -(1.0_wp/2.0_wp)*cutoff6*switchdist             &
                               +(3.0_wp/5.0_wp)*cutoff5*switchdist2            &
                               -(1.0_wp/4.0_wp)*cutoff4*switchdist3            &
                               +(1.0_wp/1.4e2_wp)*switchdist7)
          enefunc%dispersion_energy = factor*(eswitch-vlong)
          enefunc%dispersion_virial = -2.0_wp*factor*(-vswitch+vlong)

        else

          eswitch = enefunc%eswitch
          vswitch = enefunc%vswitch
          enefunc%dispersion_energy = factor*(eswitch-vlong)
          enefunc%dispersion_virial = -factor*(vswitch+vlong)

        end if

      else

        enefunc%dispersion_energy = factor*(eswitch-vlong)
        enefunc%dispersion_virial = -2.0_wp*factor*(-vswitch+vlong)

      end if
    end if

  end subroutine setup_enefunc_dispcorr

end module at_enefunc_mod
