!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_remd_mod
!> @brief   Replica exhcnage molecular dynamics simulation
!! @authors Takaharu Mori (TM), Norio Takase (NT)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_remd_mod

  use sp_energy_mod
  use sp_energy_str_mod
  use sp_energy_restraints_mod
  use sp_md_leapfrog_mod
  use sp_md_vverlet_mod
  use sp_dynamics_str_mod
  use sp_dynvars_str_mod
  use sp_dynvars_mod
  use sp_ensemble_str_mod
  use sp_remd_str_mod
  use sp_restraints_str_mod
  use sp_constraints_str_mod
  use sp_boundary_str_mod
  use sp_boundary_mod
  use sp_pairlist_str_mod
  use sp_pairlist_mod
  use sp_enefunc_str_mod
  use sp_enefunc_restraints_mod
  use sp_output_str_mod
  use sp_output_mod
  use sp_domain_str_mod
  use sp_communicate_mod
  use molecules_str_mod
  use fileio_rst_mod
  use fileio_control_mod
  use string_mod
  use random_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! structures
  type, public :: s_rep_info
    integer                           :: dimension          = 1
    integer                           :: exchange_period    = 100
    integer                           :: iseed              = 3141592
    integer,              allocatable :: types        (:)
    integer,              allocatable :: nreplicas    (:)
    character(MaxLine),   allocatable :: parameters   (:)
    logical,              allocatable :: cyclic_params(:)
    character(MaxLine),   allocatable :: rest_function(:)
  end type s_rep_info

  ! variables
  integer, public, parameter          :: RemdMaxNdimensions = 999
  integer, public, parameter          :: RemdMaxNumbrellas  = 999

  ! subroutines
  public  :: show_ctrl_remd
  public  :: read_ctrl_remd
  public  :: setup_remd
  private :: setup_reus
  public  :: run_remd
  private :: perform_replica_exchange
  private :: temperature_remd
  private :: pressure_remd
  private :: surface_tension_remd
  private :: restraint_remd
  private :: assign_condition
  private :: get_neighbouring_parmsetid
  private :: get_replica_temperature
  private :: compute_energy_restraints_remd

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_remd
  !> @brief        show REMD section usage
  !! @authors      NT, TM
  !! @param[in]    show_all : show all usage or not
  !! @param[in]    run_mode : run mode string : "md", "min", "remd"
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_remd(show_all, run_mode)

    ! formal arguments
    logical,                 intent(in)    :: show_all
    character(*),            intent(in)    :: run_mode


    if (show_all) then

      select case (run_mode)

      case ('remd')

        write(MsgOut,'(A)') '[REMD]'
        write(MsgOut,'(A)') 'dimension         = 1'
        write(MsgOut,'(A)') 'exchange_period   = 100'
        write(MsgOut,'(A)') 'iseed             = 3141592'
        write(MsgOut,'(A)') ''
        write(MsgOut,'(A)') 'type1             = TEMPERATURE'
        write(MsgOut,'(A)') 'nreplica1         = 4'
        write(MsgOut,'(A)') 'parameters1       = 300 310 320 330'
        write(MsgOut,'(A)') 'cyclic_params1    = NO'
        write(MsgOut,'(A)') ''
        write(MsgOut,'(A)') '# type2            = PRESSURE'
        write(MsgOut,'(A)') '# nreplica2        = 4'
        write(MsgOut,'(A)') '# parameters2      = 1 10 100 1000'
        write(MsgOut,'(A)') '# cyclic_params2   = NO'
        write(MsgOut,'(A)') '#'
        write(MsgOut,'(A)') '# type3            = GAMMA'
        write(MsgOut,'(A)') '# nreplica3        = 4'
        write(MsgOut,'(A)') '# parameters3      = 0 6 12 18'
        write(MsgOut,'(A)') '# cyclic_params3   = NO'
        write(MsgOut,'(A)') '#'
        write(MsgOut,'(A)') '# type4            = RESTRAINT'
        write(MsgOut,'(A)') '# nreplica4        = 4'
        write(MsgOut,'(A)') '# cyclic_params4   = NO'
        write(MsgOut,'(A)') '# rest_function4   = 1'
        write(MsgOut,'(A)') ' '

      end select

    else

      select case (run_mode)

      case ('remd')

        write(MsgOut,'(A)') '[REMD]'
        write(MsgOut,'(A)') 'exchange_period   = 100'
        write(MsgOut,'(A)') 'type1             = TEMPERATURE'
        write(MsgOut,'(A)') 'nreplica1         = 4'
        write(MsgOut,'(A)') 'parameters1       = 300 310 320 330'
        write(MsgOut,'(A)') ' '

      end select


    end if

    return

  end subroutine show_ctrl_remd
  
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_remd
  !> @brief        read REMD section in the control file
  !! @authors      TM
  !! @param[in]    handle   : unit number of control file
  !! @param[out]   rep_info : REPLICA section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_ctrl_remd(handle, rep_info)

    ! parameters
    character(*),            parameter     :: Section = 'Remd'

    ! formal arguments
    integer,                 intent(in)    :: handle
    type(s_rep_info),        intent(inout) :: rep_info

    ! local variables
    integer                  :: i, trep
    character(30)            :: cdim, numtmp, partmp, vartmp
    character(30)            :: cnstmp, reftmp, fnctmp, cyctmp


    ! read parameters from control file
    !
    call begin_ctrlfile_section(handle, Section)

    call read_ctrlfile_integer(handle, Section, 'dimension',       &
                               rep_info%dimension)
    call read_ctrlfile_integer(handle, Section, 'exchange_period', &
                               rep_info%exchange_period)
    call read_ctrlfile_integer(handle, Section, 'iseed',           &
                               rep_info%iseed)

    ! error check
    !
    if (rep_info%dimension > RemdMaxNdimensions .or. rep_info%dimension == 0) &
      call error_msg('Read_Ctrl_Remd> error in dimension in [REMD]')

    ! read allocatable variables
    !
    allocate(rep_info%types        (rep_info%dimension), &
             rep_info%nreplicas    (rep_info%dimension), &
             rep_info%parameters   (rep_info%dimension), &
             rep_info%cyclic_params(rep_info%dimension), &
             rep_info%rest_function(rep_info%dimension))

    rep_info%types        (1:rep_info%dimension) = RemdTemperature
    rep_info%nreplicas    (1:rep_info%dimension) = 0
    rep_info%parameters   (1:rep_info%dimension) = ''
    rep_info%cyclic_params(1:rep_info%dimension) = .false.
    rep_info%rest_function(1:rep_info%dimension) = ''

    do i = 1, rep_info%dimension
      write(cdim,'(i0)') i

      vartmp = 'type'          // cdim
      call read_ctrlfile_type   (handle, Section, vartmp,  &
                                 rep_info%types(i), RemdTypes)

      numtmp = 'nreplica'      // cdim
      call read_ctrlfile_integer(handle, Section, numtmp,  &
                                 rep_info%nreplicas(i))

      partmp = 'parameters'    // cdim
      call read_ctrlfile_string (handle, Section, partmp,  &
                                 rep_info%parameters(i))

      cyctmp = 'cyclic_params' // cdim
      call read_ctrlfile_logical(handle, Section, cyctmp,  &
                                 rep_info%cyclic_params(i))

      fnctmp = 'rest_function' // cdim
      call read_ctrlfile_string (handle, Section, fnctmp,  &
                                 rep_info%rest_function(i))

    end do


    ! error check
    !
    do i = 1, rep_info%dimension
      if (rep_info%nreplicas(i) == 0) &
        call error_msg('Read_Ctrl_Remd> error in nreplicas in [REMD]')
    end do

    trep = product(rep_info%nreplicas(1:rep_info%dimension))
    if (trep == 0) &
      call error_msg('Read_Ctrl_Remd> error in nreplicas in [REMD]')


    call end_ctrlfile_section(handle)

    ! write parameters to MsgOut
    !
    if (main_rank) then
      write(MsgOut,'(A)') 'Read_Ctrl_Remd> Replica information'
      do i = 1, rep_info%dimension

        write(MsgOut,'(A20,I10,A20,I10)') &
          '  dimension       = ', i,      &
          '  nreplica        = ', rep_info%nreplicas(i)

        if (rep_info%types(i) == RemdTemperature) then
          write(MsgOut,'(A)')   '  variable        = TEMPERATURE'
          write(MsgOut,'(A,A)') '  parameters      = ', trim(rep_info%parameters(i))
        else if (rep_info%types(i) == RemdPressure) then
          write(MsgOut,'(A)')   '  variable        = PRESSURE'
          write(MsgOut,'(A,A)') '  parameters      = ', trim(rep_info%parameters(i))
        else if (rep_info%types(i) == RemdGamma) then
          write(MsgOut,'(A)')   '  variable        = GAMMA'
          write(MsgOut,'(A,A)') '  parameters      = ', trim(rep_info%parameters(i))
        else if (rep_info%types(i) == RemdRestraint) then
          write(MsgOut,'(A)')   '  type            = RESTRAINT'
          write(MsgOut,'(A,A)') '  rest_function   = ', trim(rep_info%rest_function(i))
        end if

        if (rep_info%cyclic_params(i)) then
          write(MsgOut,'(A)') '  cyclic_params   = yes'
        else
          write(MsgOut,'(A)') '  cyclic_params   = no'
        end if

        write(MsgOut,'(A)') ''
      end do

    end if

    return

  end subroutine read_ctrl_remd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_remd
  !> @brief        setup REMD
  !! @authors      TM
  !! @param[in]    rep_info   : REMD section control parameters information
  !! @param[in]    rst        : Restart data
  !! @param[in]    boundary   : boundary condition information
  !! @param[in]    dynamics   : dynamics information
  !! @param[in]    molecule   : molecule information
  !! @param[inout] restraints : restraints information
  !! @param[inout] ensemble   : ensemble information
  !! @param[inout] enefunc    : potential energy functions information
  !! @param[inout] remd       : REMD information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_remd(rep_info, rst, boundary, dynamics, molecule, domain, &
                        restraints, ensemble, enefunc, remd)

    ! formal arguments
    type(s_rep_info),        intent(in)    :: rep_info
    type(s_rst),             intent(in)    :: rst
    type(s_boundary),        intent(in)    :: boundary
    type(s_dynamics),        intent(in)    :: dynamics
    type(s_molecule),        intent(in)    :: molecule
    type(s_domain),          intent(in)    :: domain
    type(s_restraints),      intent(inout) :: restraints
    type(s_ensemble),        intent(inout) :: ensemble
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_remd),    target, intent(inout) :: remd

    ! local variables
    integer                    :: i, j, k, m
    integer                    :: max_nreplicas, cycles
    integer                    :: itmp, jtmp, ktmp
    integer                    :: comp, dsta, dend, found, dtotal
    integer                    :: replicaid, parmsetid
    integer                    :: ncycle, icycle, nlen, ixx
    character(MaxLine)         :: param
    character(10)              :: tmp, frmt, cd
    character(18)              :: frmt2
    character(3)               :: ctrep
    logical                    :: keep, double
    real(wp), allocatable      :: ddata(:)
    integer,       allocatable :: ndigit(:), parmid(:)


    if (main_rank) then
      write(MsgOut,'(A)') 'Setup_Remd> Replica information'
      write(MsgOut,'(A)') ''
    end if


    ! initialize
    !
    remd%dimension       = rep_info%dimension
    remd%total_nreplicas = product(rep_info%nreplicas(1:remd%dimension))
    max_nreplicas        = maxval (rep_info%nreplicas(1:remd%dimension))

    call alloc_remd(remd, RemdReplicas, remd%dimension,  &
                    remd%total_nreplicas, max_nreplicas)

    do i = 1, remd%dimension
      remd%nreplicas(i)     = rep_info%nreplicas(i)
      remd%types(i)         = rep_info%types(i)
      remd%cyclic_params(i) = rep_info%cyclic_params(i)
    end do

    ! check ensemble 
    do i = 1, remd%dimension
      if (remd%types(i) == RemdGamma .and.  &
          ensemble%ensemble /= EnsembleNPgT) &
         call error_msg('Setup_Remd> Ensemble should be NPgT in gamma-REMD')
    end do
    if (dynamics%integrator == IntegratorVRES .or. &
        dynamics%integrator == IntegratorPMTS ) &
        call error_msg('Setup_Remd> REMD+RESPA is not allowed')

    ! check cyclic_params
    !
    do i = 1, remd%dimension
      if (remd%cyclic_params(i)) then
        if (mod(remd%nreplicas(i),2) /= 0) &
          call error_msg('Setup_Remd> nreplicas must be even,'//&
                         ' if cyclic_params is specified')
      end if
    end do


    ! setup remd_iseed, num_criteria, and num_exchange
    !
    if (rst%rstfile_type == RstfileTypeRemd) then
      remd%iseed = rst%iseed_remd
      do i = 1, remd%total_nreplicas
        remd%repid2parmsetid(i) = rst%repid2parmsetid(i)
        do j = 1, remd%dimension
          remd%num_criteria (i,j,1:2) = rst%num_criteria (i,j,1:2)
          remd%num_exchanges(i,j,1:2) = rst%num_exchanges(i,j,1:2)
        end do
      end do
    else
      remd%iseed = rep_info%iseed
      do i = 1, remd%total_nreplicas
        remd%repid2parmsetid(i) = i
        do j = 1, remd%dimension
          remd%num_criteria (i,j,1:2) = 0
          remd%num_exchanges(i,j,1:2) = 0
        end do
      end do
    end if

    ! setup permutation function
    !
    do i = 1, remd%total_nreplicas
      parmsetid = remd%repid2parmsetid(i)
      remd%parmsetid2repid(parmsetid) = i
    end do


    ! setup replica exchange period
    !
    if (rep_info%exchange_period > 0) then
      if (dynamics%eneout_period > rep_info%exchange_period) then
        call error_msg('Setup_Remd> (eneout_period <= exchange_period)'//&
                       ' is required')
      else
        if (mod(rep_info%exchange_period,dynamics%eneout_period) /= 0) then
          call error_msg('Setup_Remd> mod(exchange_period,eneout_period)'//&
                         ' must be zero)')
        end if
      end if
      cycles = 2*rep_info%exchange_period*remd%dimension
      if (mod(dynamics%nsteps,cycles) /= 0) then
        call error_msg('Setup_Remd> mod(nsteps,(2*exchange_period*dimension))'//&
                       ' must be zero')
      else
        remd%equilibration_only = .false.
        remd%exchange_period = rep_info%exchange_period
        remd%ncycles = dynamics%nsteps/remd%exchange_period
      end if
    else if (rep_info%exchange_period == 0) then
      remd%equilibration_only = .true.
      remd%exchange_period = dynamics%nsteps
      remd%ncycles = 1
    else
      call error_msg('Setup_Remd> error in exchange_period')
    end if


    ! split rep_info%parameters (char) into parameters (real)
    !
    allocate(ddata(max_nreplicas))

    do i = 1, remd%dimension
      if (remd%types(i) /= RemdRestraint) then
        param = rep_info%parameters(i)
        if (split_num(param) /= remd%nreplicas(i)) &
          call error_msg('Setup_Remd> "nreplicas" and number of data in "parameters" are inconsistent')

        call split(remd%nreplicas(i), max_nreplicas, param, ddata)

        do j = 1, remd%nreplicas(i)
          remd%dparameters(i,j) = ddata(j)
        end do
      end if
    end do

    deallocate(ddata)


    ! setup replica-exchange umbrella sampling (REUS)
    !
    call setup_reus(rep_info, restraints, enefunc, remd)


    ! generate multi-dimensional parameter sets
    !
    allocate(ndigit(remd%dimension),parmid(remd%dimension))

    do i = 1, remd%dimension
      do j = 1, 100
        comp = 10**j
        if (remd%nreplicas(i) < comp) then
          ndigit(i) = j
          exit
        end if
      end do
    end do
    dtotal = sum(ndigit(1:remd%dimension))

    write(cd,'(i10)') dtotal
    frmt = '(i' // trim(adjustl(cd)) // '.' // trim(adjustl(cd)) // ')'

    found     = 0
    parmsetid = 0
    do
      found = found + 1
      write(tmp,frmt) found
      keep = .true.
      dend = 0

      do j = 1, remd%dimension
        dsta = dend + 1
        dend = dsta + ndigit(j) - 1
        read(tmp(dsta:dend),*) itmp
        if (itmp > remd%nreplicas(j) .or. itmp == 0) then
          keep = .false.
          exit
        end if
      end do

      if (keep) then
        parmsetid = parmsetid + 1
        dend = 0
        do j = 1, remd%dimension
          dsta = dend + 1
          dend = dsta + ndigit(j) - 1
          read(tmp(dsta:dend),*) parmid(j)
          remd%parmidsets(parmsetid,j) = parmid(j)
        end do
        if (parmsetid == remd%total_nreplicas) exit
      end if
    end do

    deallocate(ndigit,parmid)


    ! assign itnitial parameters
    !
    replicaid = my_country_no + 1
    parmsetid = remd%repid2parmsetid(replicaid)
    call assign_condition(parmsetid, remd, domain, ensemble, enefunc)


    ! write the summary of setup
    !
    if (main_rank) then
      write(MsgOut,'(A)') '  ParmsetID'

      write(ctrep,'(I3)') remd%total_nreplicas
      frmt2 = '(I11 , A,' // ctrep // 'I11  )'

      do i = 1, remd%total_nreplicas
        replicaid = i
        parmsetid = remd%repid2parmsetid(replicaid)
        write(MsgOut,frmt2) parmsetid, ' = ', &
          remd%parmidsets(parmsetid,1:remd%dimension)
      end do

      write(MsgOut,'(A)') ''
      write(MsgOut,'(A)') '  Parameters'

      do i = 1, remd%dimension
        if (remd%types(i) /= RemdRestraint) then
          write(ctrep,'(I3)') remd%nreplicas(i)
          frmt2 = '(A,I4,A,' // ctrep // 'F11.3)'
          write(MsgOut,frmt2) &
            '    Dim', i, ' = ', remd%dparameters(i,1:remd%nreplicas(i))
        else
          write(ctrep,'(I3)') remd%nreplicas(i)
          frmt2 = '(A,I4,A,' // ctrep // 'I11  )'
          write(MsgOut,frmt2) &
            '    Dim', i, ' = ', remd%iparameters(i,1:remd%nreplicas(i))
        end if
      end do

      do i = 1, remd%dimension
        if (remd%types(i) == RemdRestraint) then
          write(MsgOut,'(A)') ''
          write(MsgOut,'(A)') '  Restraints'
          exit
        end if
      end do

      do i = 1, size(remd%umbrid2numfuncs(:))
        write(MsgOut,'(4X,A,I10)') 'umbrella potential index = ', i
          do k = 1, remd%umbrid2numfuncs(i)
            write(MsgOut,'(6X,A,I7,2X,A,F11.3,2X,A,F11.3)')         &
                         'function = ',  remd%umbrid2funclist(i,k), &
                         'constant = ',  remd%rest_constants(i,k),  &
                         'reference = ', remd%rest_reference(i,k)
          end do
        write(MsgOut,'(A)') ''
      end do

    end if

    return

  end subroutine setup_remd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_reus
  !> @brief        setup REUS
  !! @authors      TM
  !! @param[in]    rep_info   : REMD section control parameters information
  !! @param[in]    rst        : Restart data
  !! @param[in]    boundary   : boundary condition information
  !! @param[in]    dynamics   : dynamics information
  !! @param[in]    molecule   : molecule information
  !! @param[inout] restraints : restraints information
  !! @param[inout] ensemble   : ensemble information
  !! @param[inout] enefunc    : potential energy functions information
  !! @param[inout] remd       : REMD information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_reus(rep_info, restraints, enefunc, remd)

    ! formal arguments
    type(s_rep_info),        intent(in)    :: rep_info
    type(s_restraints),      intent(inout) :: restraints
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_remd),    target, intent(inout) :: remd

    ! local variables
    integer                    :: i, j, k, ifound
    integer                    :: ndata, max_ndata
    integer                    :: umbrid, funcid, max_umbrella_id, max_nreplicas
    character(MaxLine)         :: param
    character(10)              :: msg1, msg2
    integer,       allocatable :: idata(:)
    real(wp),      allocatable :: ddata(:)


    ! count numbers for allocation
    !
    max_nreplicas = maxval(rep_info%nreplicas(1:remd%dimension))
    max_ndata = 0
    ifound = 0
    do i = 1, remd%dimension
      if (remd%types(i) == RemdRestraint) then
        param = rep_info%rest_function(i)
        ndata = split_num(param)
        if (ndata > max_ndata) max_ndata = ndata
        ifound = ifound + remd%nreplicas(i)
      end if
    end do
    max_umbrella_id = ifound

    call alloc_remd(remd, RemdUmbrellas, max_umbrella_id, max_ndata, max_nreplicas)


    ! split rep_info%rest_function (char) into parameters (integer)
    !   and make permutation function for REUS
    !
    allocate(idata(max_ndata))

    umbrid = 0
    do i = 1, remd%dimension
      if (remd%types(i) == RemdRestraint) then
        param = rep_info%rest_function(i)
        ndata = split_num(param)

        call split(ndata, max_ndata, param, idata)

        do j = 1, remd%nreplicas(i)
          umbrid = umbrid + 1
          remd%iparameters(i,j) = umbrid
          remd%umbrid2numfuncs(umbrid) = ndata
          remd%umbrid2funclist(umbrid,1:ndata) = idata(1:ndata)
        end do
      end if
    end do

    deallocate(idata)


    ! check consistency between [REMD] and [RESTRAINTS]
    !
    umbrid = 0
    do i = 1, remd%dimension
      if (remd%types(i) == RemdRestraint) then
        do j = 1, remd%nreplicas(i)
          umbrid = umbrid + 1
          do k = 1, remd%umbrid2numfuncs(umbrid)
            funcid = remd%umbrid2funclist(umbrid,k)

            write(msg1,'(i0)') i
            write(msg2,'(i0)') funcid

            if (enefunc%restraint_kind(funcid) == RestraintsFuncPOSI) then
              call error_msg('Setup_Remd> REUS of positional restraints is not allowed')
            end if
            if (enefunc%restraint_kind(funcid) == RestraintsFuncPC .or.   &
                enefunc%restraint_kind(funcid) == RestraintsFuncPCCOM) then
              call error_msg('Setup_Remd> REUS of PC restraints is not allowed')
            end if
            if (split_num(restraints%constant(funcid)) /= remd%nreplicas(i)) then
              call error_msg('Setup_Remd> nreplica' // trim(msg1) // ' in [REMD] and &
                                          constant' // trim(msg2) // ' in [RESTRAINTS] &
                                          have inconsistency')
            end if
            if (split_num(restraints%reference(funcid)) /= remd%nreplicas(i)) then
              call error_msg('Setup_Remd> nreplica' // trim(msg1) // ' in [REMD] and &
                                          reference'// trim(msg2) // ' in [RESTRAINTS] &
                                          have inconsistency')
            end if
            if (funcid > restraints%nfunctions .or. funcid <= 0) then
              call error_msg('Setup_Remd> Cannot define index = ' // trim(msg2) // &
                                          ' in rest_function' // trim(msg1) // ' in [REMD]')
            end if

          end do
        end do
      end if
    end do


    ! setup constants and reference in umbrella potentials
    !
    allocate(ddata(max_nreplicas))

    umbrid = 0
    do i = 1, remd%dimension
      if (remd%types(i) == RemdRestraint) then
        do j = 1, remd%nreplicas(i)
          umbrid = umbrid + 1
          do k = 1, remd%umbrid2numfuncs(umbrid)
            funcid = remd%umbrid2funclist(umbrid,k)

            ! force constant
            param = restraints%constant(funcid)
            ndata = split_num(param)
            call split(ndata, max_nreplicas, param, ddata)
            remd%rest_constants(umbrid,k) = ddata(j)

            ! reference value
            param = restraints%reference(funcid)
            ndata = split_num(param)
            call split(ndata, max_nreplicas, param, ddata)
            remd%rest_reference(umbrid,k) = ddata(j)

          end do
        end do
      end if
    end do

    deallocate(ddata)

    return

  end subroutine setup_reus

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    run_remd
  !> @brief        control replica exchange
  !! @authors      TM
  !! @param[inout] output      : output information
  !! @param[inout] molecule    : molecule information
  !! @param[inout] enefunc     : potential energy functions information
  !! @param[inout] dynvars     : dynamics variables information
  !! @param[inout] dynamics    : dynamics information
  !! @param[inout] pairlist    : pairlist information
  !! @param[inout] boundary    : boundary conditions information
  !! @param[inout] constraints : constraints information
  !! @param[inout] ensemble    : ensemble information
  !! @param[inout] remd        : REMD information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine run_remd(output, domain, enefunc, dynvars, dynamics, &
                      pairlist, boundary, constraints, ensemble,  &
                      comm, remd)

    ! formal arguments
    type(s_output),          intent(inout) :: output
    type(s_domain),  target, intent(inout) :: domain
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_dynvars),         intent(inout) :: dynvars
    type(s_dynamics),        intent(inout) :: dynamics
    type(s_pairlist),        intent(inout) :: pairlist
    type(s_boundary),        intent(inout) :: boundary
    type(s_constraints),     intent(inout) :: constraints
    type(s_ensemble),        intent(inout) :: ensemble
    type(s_comm),            intent(inout) :: comm
    type(s_remd),            intent(inout) :: remd

    ! local variables
    integer                   :: i, replicaid, parmsetid


    ! Open output files
    !
    call open_output(output)

    ! output restart data
    !
    dynvars%step = 0
    call output_remd(0, output, domain, dynamics, dynvars, boundary, remd)

    do i = 1, remd%ncycles

      dynamics%istart_step  = (i-1)*remd%exchange_period + 1
      dynamics%iend_step    =  i   *remd%exchange_period
      dynamics%initial_time = dynvars%time

      ! set conditions
      !
      replicaid = my_country_no + 1
      parmsetid = remd%repid2parmsetid(replicaid)
      call assign_condition(parmsetid, remd, domain, ensemble, enefunc)

      ! MD main loop
      !
      if (dynamics%integrator == IntegratorLEAP) then
        call leapfrog_dynamics(output, domain, enefunc, dynvars, dynamics, &
                               pairlist, boundary, constraints, ensemble,  &
                               comm)
      else if (dynamics%integrator == IntegratorVVER) then
        call vverlet_dynamics (output, domain, enefunc, dynvars, dynamics, &
                               pairlist, boundary, constraints, ensemble,  &
                               comm)
      end if

      ! perform remd
      !
      if (.not. remd%equilibration_only) then
        call perform_replica_exchange(i, domain, enefunc, dynvars,         &
                                      ensemble, boundary, output, remd)
      end if

      ! output restart data
      !
      call output_remd(i, output, domain, dynamics, dynvars, boundary, remd)

    end do

    ! close output files
    !
    call close_output(output)


    return

  end subroutine run_remd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    perform_replica_exchange
  !> @brief        perform replica exchange
  !! @authors      TM
  !! @param[inout] molecule    : molecule information
  !! @param[inout] enefunc     : potential energy functions information
  !! @param[inout] dynvars     : dynamics variables information
  !! @param[inout] boundary    : boundary conditions information
  !! @param[inout] ensemble    : ensemble information
  !! @param[inout] remd        : REMD information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine perform_replica_exchange(icycle, domain, enefunc, dynvars, &
                                      ensemble, boundary, output, remd)

    ! formal arguments
    integer,                  intent(in)    :: icycle
    type(s_domain),   target, intent(inout) :: domain
    type(s_enefunc),          intent(inout) :: enefunc
    type(s_dynvars),  target, intent(inout) :: dynvars
    type(s_ensemble),         intent(inout) :: ensemble
    type(s_boundary),         intent(inout) :: boundary
    type(s_output),           intent(inout) :: output
    type(s_remd),     target, intent(inout) :: remd

    ! local variables
    integer                   :: exptrn, dimno, itmp, parmsetid
    integer                   :: i, j, neighid, id_old
    integer                   :: ipara, ipara_ref
    real(wp)                  :: dpara, dpara_ref
    character                 :: ctrep*3, frmt1*12, frmt2*12, aorr*1
    integer,          pointer :: total_nreplicas, iseed
    integer,          pointer :: num_criteria(:,:,:), num_exchanges(:,:,:)
    integer,          pointer :: parmidsets(:,:)
    integer,          pointer :: repid2parmsetid(:), parmsetid2repid(:)
    integer,          pointer :: repid2parmsetid_ref(:)
    integer,          pointer :: iparameters(:,:)
    real(wp),         pointer :: dparameters(:,:)


    iseed               => remd%iseed
    total_nreplicas     => remd%total_nreplicas
    repid2parmsetid     => remd%repid2parmsetid
    repid2parmsetid_ref => remd%repid2parmsetid_ref
    parmsetid2repid     => remd%parmsetid2repid
    parmidsets          => remd%parmidsets
    dparameters         => remd%dparameters
    iparameters         => remd%iparameters
    num_criteria        => remd%num_criteria
    num_exchanges       => remd%num_exchanges


    ! decide dimension to be exchanged and exchange pattern
    !   pattern 1: 1<>2  3<>4  5<>6  7<>8 ...
    !   pattern 2: 1  2<>3  4<>5  6<>7  8 ...
    !
    itmp  = mod(icycle,2*remd%dimension)
    dimno = mod(itmp,remd%dimension)+1
    if (itmp+1 <= remd%dimension) then
      exptrn = 1
    else
      exptrn = 2
    end if


    ! save previous information
    !
    repid2parmsetid_ref(1:total_nreplicas) = repid2parmsetid(1:total_nreplicas)


    ! generate random number
    !
    do i = 1, total_nreplicas-1
      do j = i+1, total_nreplicas
        remd%random_table(i,j) = random_get_legacy(iseed)
        remd%random_table(j,i) = remd%random_table(i,j)
      end do
    end do

    
    ! perform replica exchange
    !
    select case(remd%types(dimno))

    case(RemdTemperature)

      call temperature_remd(dimno, exptrn, domain, dynvars, remd)

    case(RemdPressure)

      call pressure_remd(dimno, exptrn, ensemble, boundary, remd)

    case(RemdGamma)

      call surface_tension_remd(dimno, exptrn, ensemble, boundary, remd)

    case(RemdRestraint)

      call restraint_remd(dimno, exptrn, domain, ensemble, dynvars,      &
                          enefunc, remd)

    end select


    ! update permutation function
    !
    do i = 1, total_nreplicas
      parmsetid = repid2parmsetid(i)
      parmsetid2repid(parmsetid) = i
    end do


    ! write results
    !
    if (main_rank) then

      write(MsgOut,'(A,I10,3X,A,I4,3X,A,I4)')                           &
        'REMD> Step: ',dynvars%step,'Dimension: ',dimno,                &
        'ExchangePattern: ',exptrn

      write(MsgOut,'(A,6X,A,13X,A,6X,A,7X,A)')     &
        '  Replica', 'ExchangeTrial', 'AcceptanceRatio', 'Before', 'After'
      do i = 1, total_nreplicas
        id_old = repid2parmsetid_ref(i)
        call get_neighbouring_parmsetid(dimno, exptrn, remd, id_old, neighid)

        if (remd%types(dimno) /= RemdRestraint) then

          dpara_ref = dparameters(dimno,parmidsets(repid2parmsetid_ref(i),dimno))
          dpara     = dparameters(dimno,parmidsets(repid2parmsetid(i),    dimno))

          if (abs(dpara_ref - dpara) < EPS) then
            aorr = 'R'
          else
            aorr = 'A'
          end if
          if (neighid == 0) aorr = 'N'

          write(MsgOut,'(I9,6X,I5,A,I5,3X,A1,3X,I9,A,I9,2F12.3)')         &
               i, id_old, ' > ', neighid, aorr,                           &
               num_exchanges(repid2parmsetid_ref(i),dimno,exptrn), ' / ', &
               num_criteria (repid2parmsetid_ref(i),dimno,exptrn),        &
               dpara_ref, dpara
        else

          ipara_ref = iparameters(dimno,parmidsets(repid2parmsetid_ref(i),dimno))
          ipara     = iparameters(dimno,parmidsets(repid2parmsetid(i),    dimno))

          if (ipara_ref == ipara) then
            aorr = 'R'
          else
            aorr = 'A'
          end if
          if (neighid == 0) aorr = 'N'

          write(MsgOut,'(I9,6X,I5,A,I5,3X,A1,3X,I9,A,I9,2I12)')           &
               i, id_old, ' > ', neighid, aorr,                           &
               num_exchanges(repid2parmsetid_ref(i),dimno,exptrn), ' / ', &
               num_criteria (repid2parmsetid_ref(i),dimno,exptrn),        &
               ipara_ref, ipara
        end if
      end do

      write(MsgOut,*) ''

      write(ctrep,'(I3)') total_nreplicas
      frmt1 = '(A,' // trim(adjustl(ctrep)) // 'F10.3)'
      frmt2 = '(A,' // trim(adjustl(ctrep)) // 'I10  )'

      if (remd%types(dimno) /= RemdRestraint) then
        write(MsgOut,frmt1) '  Parameter    : ', &
          dparameters(dimno,parmidsets(repid2parmsetid(1:total_nreplicas),dimno))
      else
        write(MsgOut,frmt2) '  Parameter    : ', &
          iparameters(dimno,parmidsets(repid2parmsetid(1:total_nreplicas),dimno))
      end if
      write(MsgOut,frmt2)   '  RepIDtoParmID: ', repid2parmsetid(1:total_nreplicas)
      write(MsgOut,frmt2)   '  ParmIDtoRepID: ', parmsetid2repid(1:total_nreplicas)

      write(MsgOut,*) ''

    end if

    return

  end subroutine perform_replica_exchange

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    temperature_remd
  !> @brief        temperature REMD
  !! @authors      TM
  !! @param[in]    dimno    : dimension number
  !! @param[in]    exptrn   : exchange pattern
  !! @param[inout] molecule : molecule information
  !! @param[inout] remd     : REMD information
  !! @note         Y.Sugita & Y.Okamoto, Chem.Phys.Lett., 314, 141−151 (1999)
  !!               K.Hukushima & K.Nemoto, J.Phys.Soc.Jpn., 65, 1604-1608 (1996)
  !!               Y.Mori & Y.Okamoto, J.Phys.Soc.Jpn., 79, 074001 (2010)
  !!               Y.Mori & Y.Okamoto, J.Phys.Soc.Jpn., 79, 074003 (2010)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine temperature_remd(dimno, exptrn, domain, dynvars, remd)

    ! formal arguments
    integer,                  intent(in)    :: dimno
    integer,                  intent(in)    :: exptrn
    type(s_domain),   target, intent(inout) :: domain
    type(s_dynvars),          intent(inout) :: dynvars
    type(s_remd),     target, intent(inout) :: remd

    ! local variables
    integer                   :: i, j, jx
    integer                   :: repid_i, repid_j, neighid
    integer                   :: parmsetid, parmsetid_i, parmsetid_j
    real(wp)                  :: temp_i, temp_j, beta_i, beta_j, factor
    real(wp)                  :: energy_i, energy_j, delta
    real(wp)                  :: before_gather, eneval
    logical                   :: do_trial, do_exchange
    integer,          pointer :: total_nreplicas, iseed
    integer,          pointer :: num_criteria(:,:,:), num_exchanges(:,:,:)
    integer,          pointer :: parmidsets(:,:)
    integer,          pointer :: repid2parmsetid(:), parmsetid2repid(:)
    real(wp),         pointer :: parameters(:,:)
    real(wp),         pointer :: after_gather(:), random_table(:,:)


    total_nreplicas => remd%total_nreplicas
    repid2parmsetid => remd%repid2parmsetid
    parmsetid2repid => remd%parmsetid2repid
    parmidsets      => remd%parmidsets
    parameters      => remd%dparameters
    num_criteria    => remd%num_criteria
    num_exchanges   => remd%num_exchanges
    iseed           => remd%iseed
    after_gather    => remd%after_gather
    random_table    => remd%random_table


    ! allgather potential energy
    !
    eneval = dynvars%energy%total
    call mpi_bcast(eneval, 1, mpi_wp_real, 0, mpi_comm_country, ierror)

    before_gather = eneval
#ifdef MPI
    call mpi_allgather(before_gather, 1, mpi_wp_real, &
                       after_gather,  1, mpi_wp_real, &
                       mpi_comm_airplane, ierror)
#endif
    do i = 1, total_nreplicas
      remd%potential_energy(i) = after_gather(i)
    end do

    ! replica exchange
    !
    do i = 1, total_nreplicas

      repid_i     = i
      parmsetid_i = repid2parmsetid(repid_i)

      ! get neighboring parameter set index and replica index
      !
      call get_neighbouring_parmsetid(dimno, exptrn, remd, parmsetid_i, neighid)

      if (neighid /= 0) then
        parmsetid_j = neighid
        repid_j     = parmsetid2repid(parmsetid_j)
        do_trial    = .true.
      else
        do_trial    = .false.
      end if


      if (do_trial) then

        num_criteria(parmsetid_i,dimno,exptrn) &
          = num_criteria(parmsetid_i,dimno,exptrn) + 1

        ! compute delta
        !
        energy_i = remd%potential_energy(repid_i)
        energy_j = remd%potential_energy(repid_j)
        temp_i   = parameters(dimno,parmidsets(parmsetid_i,dimno))
        temp_j   = parameters(dimno,parmidsets(parmsetid_j,dimno))
        beta_i   = 1.0_wp / (KBOLTZ * temp_i)
        beta_j   = 1.0_wp / (KBOLTZ * temp_j)
        delta    = (beta_j - beta_i)*(energy_i - energy_j)

        ! Metropolis criterion
        !
        if (delta <= 0.0_wp) then
          do_exchange = .true.
        else if (delta > 0.0_wp) then
          if (random_table(repid_i,repid_j) <= exp(-delta)) then
            do_exchange = .true.
          else
            do_exchange = .false.
          end if
        end if


        if (do_exchange) then

          num_exchanges(parmsetid_i,dimno,exptrn) &
            = num_exchanges(parmsetid_i,dimno,exptrn) + 1

          ! scale velocities and themostat & barostat momenta
          !
          if (i == my_country_no + 1) then

            factor = sqrt(temp_j/temp_i)
            do j = 1, domain%num_cell_local
              do jx = 1, domain%num_atom(j)
                domain%velocity(1:3,jx,j) = domain%velocity(1:3,jx,j) * factor
              end do
            end do

            factor = temp_j/temp_i
            dynvars%thermostat_momentum                 &
              = dynvars%thermostat_momentum * factor
            dynvars%barostat_momentum(1:3)              &
              = dynvars%barostat_momentum(1:3) * factor

          end if

          repid2parmsetid(i) = parmsetid_j
        end if
 
      end if

    end do

    return

  end subroutine temperature_remd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pressure_remd
  !> @brief        pressure REMD
  !! @authors      TM
  !! @param[in]    dimno    : dimension number
  !! @param[in]    exptrn   : exchange pattern
  !! @param[in]    ensemble : ensemble information
  !! @param[inout] remd     : REMD information
  !! @note         T.Okabe et al., Chem.Phys.Lett., 335, 435−439 (2001)
  !!               Y.Mori & Y.Okamoto, J.Phys.Soc.Jpn., 79, 074003 (2010)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pressure_remd(dimno, exptrn, ensemble, boundary, remd)

    ! formal arguments
    integer,                 intent(in)    :: dimno
    integer,                 intent(in)    :: exptrn
    type(s_ensemble),        intent(in)    :: ensemble
    type(s_boundary),        intent(in)    :: boundary
    type(s_remd),    target, intent(inout) :: remd

    ! local variables
    integer                  :: i, j, k
    integer                  :: repid_i, repid_j, neighid
    integer                  :: parmsetid, parmsetid_i, parmsetid_j
    real(wp)                 :: beta, delta, temperature
    real(wp)                 :: volume_i, volume_j, press_i, press_j
    real(wp)                 :: before_gather
    logical                  :: do_trial, do_exchange
    integer,         pointer :: total_nreplicas
    integer,         pointer :: num_criteria(:,:,:), num_exchanges(:,:,:)
    integer,         pointer :: parmidsets(:,:)
    integer,         pointer :: repid2parmsetid(:), parmsetid2repid(:)
    real(wp),        pointer :: parameters(:,:)
    real(wp),        pointer :: after_gather(:), random_table(:,:)


    total_nreplicas => remd%total_nreplicas
    repid2parmsetid => remd%repid2parmsetid
    parmsetid2repid => remd%parmsetid2repid
    parmidsets      => remd%parmidsets
    parameters      => remd%dparameters
    num_criteria    => remd%num_criteria
    num_exchanges   => remd%num_exchanges
    after_gather    => remd%after_gather
    random_table    => remd%random_table


    ! allgather volume
    !
    before_gather =  boundary%box_size_x &
                   * boundary%box_size_y &
                   * boundary%box_size_z 
#ifdef MPI
    call mpi_allgather(before_gather, 1, mpi_wp_real, &
                       after_gather,  1, mpi_wp_real, &
                       mpi_comm_airplane, ierror)
#endif
    do i = 1, total_nreplicas
      remd%volume(i) = after_gather(i)
    end do

    ! replica exchange
    !
    do i = 1, total_nreplicas

      repid_i     = i
      parmsetid_i = repid2parmsetid(repid_i)

      ! get neighboring parameter set index and replica index
      !
      call get_neighbouring_parmsetid(dimno, exptrn, remd, parmsetid_i, neighid)

      if (neighid /= 0) then
        parmsetid_j = neighid
        repid_j     = parmsetid2repid(parmsetid_j)
        do_trial    = .true.
      else
        do_trial    = .false.
      end if


      if (do_trial) then

        num_criteria(parmsetid_i,dimno,exptrn) &
          = num_criteria(parmsetid_i,dimno,exptrn) + 1

        ! compute delta
        !
        call get_replica_temperature(parmsetid_i, remd, ensemble, temperature)

        volume_i = remd%volume(repid_i)
        volume_j = remd%volume(repid_j)
        beta     = 1.0_wp / (KBOLTZ * temperature)
        press_i  = parameters(dimno,parmidsets(parmsetid_i,dimno))*ATMOS_P
        press_j  = parameters(dimno,parmidsets(parmsetid_j,dimno))*ATMOS_P
        delta    = beta*(press_i - press_j)*(volume_j - volume_i)

        ! Metropolis criterion
        !
        if (delta <= 0.0_wp) then
          do_exchange = .true.
        else if (delta > 0.0_wp) then
          if (random_table(repid_i,repid_j) <= exp(-delta)) then
            do_exchange = .true.
          else
            do_exchange = .false.
          end if
        end if

        if (do_exchange) then
          num_exchanges(parmsetid_i,dimno,exptrn) &
            = num_exchanges(parmsetid_i,dimno,exptrn) + 1

          repid2parmsetid(i) = parmsetid_j
        end if
 
      end if

    end do

    return

  end subroutine pressure_remd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    surface_tension_remd
  !> @brief        surface-tension REMD
  !! @authors      TM
  !! @param[in]    dimno    : dimension number
  !! @param[in]    exptrn   : exchange pattern
  !! @param[in]    ensemble : ensemble information
  !! @param[inout] remd     : REMD information
  !! @note         T.Mori et al., J.Chem.Theory.Comput., 9, 5629−5640 (2013)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine surface_tension_remd(dimno, exptrn, ensemble, boundary, remd)

    ! formal arguments
    integer,                 intent(in)    :: dimno
    integer,                 intent(in)    :: exptrn
    type(s_ensemble),        intent(in)    :: ensemble
    type(s_boundary),        intent(in)    :: boundary
    type(s_remd),    target, intent(inout) :: remd

    ! local variables
    integer                  :: i, j, k
    integer                  :: repid_i, repid_j, neighid
    integer                  :: parmsetid, parmsetid_i, parmsetid_j
    real(wp)                 :: beta, factor, delta, temperature
    real(wp)                 :: area_i, area_j, gamma_i, gamma_j
    real(wp)                 :: before_gather
    logical                  :: do_trial, do_exchange
    integer,         pointer :: total_nreplicas
    integer,         pointer :: num_criteria(:,:,:), num_exchanges(:,:,:)
    integer,         pointer :: parmidsets(:,:)
    integer,         pointer :: repid2parmsetid(:), parmsetid2repid(:)
    real(wp),        pointer :: parameters(:,:)
    real(wp),        pointer :: after_gather(:), random_table(:,:)


    total_nreplicas => remd%total_nreplicas
    repid2parmsetid => remd%repid2parmsetid
    parmsetid2repid => remd%parmsetid2repid
    parmidsets      => remd%parmidsets
    parameters      => remd%dparameters
    num_criteria    => remd%num_criteria
    num_exchanges   => remd%num_exchanges
    after_gather    => remd%after_gather
    random_table    => remd%random_table


    ! allgather area
    !
    before_gather =  boundary%box_size_x &
                   * boundary%box_size_y 
#ifdef MPI
    call mpi_allgather(before_gather, 1, mpi_wp_real, &
                       after_gather,  1, mpi_wp_real, &
                       mpi_comm_airplane, ierror)
#endif
    do i = 1, total_nreplicas
      remd%area(i) = after_gather(i)
    end do

    ! replica exchange
    !
    do i = 1, total_nreplicas

      repid_i     = i
      parmsetid_i = repid2parmsetid(repid_i)

      ! get neighboring parameter set index and replica index
      !
      call get_neighbouring_parmsetid(dimno, exptrn, remd, parmsetid_i, neighid)

      if (neighid /= 0) then
        parmsetid_j = neighid
        repid_j     = parmsetid2repid(parmsetid_j)
        do_trial    = .true.
      else
        do_trial    = .false.
      end if


      if (do_trial) then

        num_criteria(parmsetid_i,dimno,exptrn) &
          = num_criteria(parmsetid_i,dimno,exptrn) + 1

        ! compute delta
        !
        call get_replica_temperature(parmsetid_i, remd, ensemble, temperature)

        factor  = ATMOS_P*100.0_wp/1.01325_wp
        area_i  = remd%area(repid_i)
        area_j  = remd%area(repid_j)
        gamma_i = parameters(dimno,parmidsets(parmsetid_i,dimno))*factor
        gamma_j = parameters(dimno,parmidsets(parmsetid_j,dimno))*factor
        beta    = 1.0_wp / (KBOLTZ * temperature)
        delta   = -beta*(gamma_i - gamma_j)*(area_j - area_i)

        ! Metropolis criterion
        !
        if (delta <= 0.0_wp) then
          do_exchange = .true.
        else if (delta > 0.0_wp) then
          if (random_table(repid_i,repid_j) <= exp(-delta)) then
            do_exchange = .true.
          else
            do_exchange = .false.
          end if
        end if

        if (do_exchange) then
          num_exchanges(parmsetid_i,dimno,exptrn) &
            = num_exchanges(parmsetid_i,dimno,exptrn) + 1

          repid2parmsetid(i) = parmsetid_j
        end if
 
      end if

    end do

    return

  end subroutine surface_tension_remd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    restraint_remd
  !> @brief        Replica exchange umbrella sampling (REUS)
  !! @authors      TM
  !! @param[in]    dimno    : dimension number
  !! @param[in]    exptrn   : exchange pattern
  !! @param[inout] enefunc  : potential energy functions information
  !! @param[inout] remd     : REMD information
  !! @param[inout] ensemble : ensemble information
  !! @param[inout] dynvars  : dynamics variables information
  !! @note         Y.Sugita et al., J.Chem.Phys., 113, 6042-6051 (2000)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine restraint_remd(dimno, exptrn, domain, ensemble, dynvars, &
                            enefunc, remd)

    ! formal arguments
    integer,                  intent(in)    :: dimno
    integer,                  intent(in)    :: exptrn
    type(s_domain),   target, intent(inout) :: domain
    type(s_ensemble),         intent(in)    :: ensemble
    type(s_dynvars),  target, intent(inout) :: dynvars
    type(s_enefunc),  target, intent(inout) :: enefunc
    type(s_remd),     target, intent(inout) :: remd

    ! local variables
    integer                  :: i, j, k, ix
    integer                  :: natoms
    integer                  :: repid_i, repid_j
    integer                  :: parmsetid, parmsetid_i, parmsetid_j
    integer                  :: neighid
    integer                  :: funcid, umbr_0, umbr_1
    real(wp)                 :: cv
    real(wp)                 :: energy_i_0, energy_j_0
    real(wp)                 :: energy_i_1, energy_j_1
    real(wp)                 :: eneval, beta, delta, temperature
    real(wp)                 :: before_gather
    logical                  :: do_trial, do_exchange, umbr_exist
    integer,         pointer :: total_nreplicas
    integer,         pointer :: num_criteria(:,:,:), num_exchanges(:,:,:)
    integer,         pointer :: parmidsets(:,:)
    integer,         pointer :: repid2parmsetid(:), parmsetid2repid(:)
    integer,         pointer :: parameters(:,:)
    real(wp),        pointer :: after_gather(:), random_table(:,:)
    real(wp),        pointer :: refcoord0(:,:), refcoord1(:,:)


    after_gather    => remd%after_gather
    random_table    => remd%random_table
    total_nreplicas => remd%total_nreplicas
    repid2parmsetid => remd%repid2parmsetid
    parmsetid2repid => remd%parmsetid2repid
    parmidsets      => remd%parmidsets
    parameters      => remd%iparameters
    num_criteria    => remd%num_criteria
    num_exchanges   => remd%num_exchanges
    refcoord0       => remd%restraint_refcoord0
    refcoord1       => remd%restraint_refcoord1

    ! set target restraint function
    !
    repid_i     = my_country_no + 1
    parmsetid_i = repid2parmsetid(repid_i)
    umbr_0      = parameters(dimno,parmidsets(parmsetid_i,dimno))


    ! calculate restraint potential energy at the current condition
    !
    do k = 1, remd%umbrid2numfuncs(umbr_0)
      funcid = remd%umbrid2funclist(umbr_0,k)
      enefunc%restraint_const(1:2,funcid) = remd%rest_constants(umbr_0,k)
      enefunc%restraint_ref  (1:2,funcid) = remd%rest_reference(umbr_0,k)

      if (enefunc%restraint_kind(funcid) == RestraintsFuncRMSD) then
        enefunc%rmsd_force  = enefunc%restraint_const(1,funcid)
        enefunc%rmsd_target = enefunc%restraint_ref  (1,funcid)
      end if
    end do

    call compute_energy_restraints_remd(.true., enefunc, domain, eneval)
    before_gather = eneval

#ifdef MPI
    call mpi_allgather(before_gather, 1, mpi_wp_real, &
                       after_gather,  1, mpi_wp_real, &
                       mpi_comm_airplane, ierror)
#endif
    do i = 1, total_nreplicas
      remd%restraint_energy_0(i) = after_gather(i)
    end do


    ! calculate restraint potential energy using the neighboring condition
    !
    call get_neighbouring_parmsetid(dimno, exptrn, remd, parmsetid_i, neighid)

    if (neighid /= 0) then
      parmsetid_j = neighid
      repid_j     = parmsetid2repid(parmsetid_j)
      umbr_exist  = .true.
    else
      umbr_exist  = .false.
    end if

    if (umbr_exist) then
      umbr_1 = parameters(dimno,parmidsets(parmsetid_j,dimno))

      do k = 1, remd%umbrid2numfuncs(umbr_1)
        funcid = remd%umbrid2funclist(umbr_1,k)
        enefunc%restraint_const(1:2,funcid) = remd%rest_constants(umbr_1,k)
        enefunc%restraint_ref  (1:2,funcid) = remd%rest_reference(umbr_1,k)

        if (enefunc%restraint_kind(funcid) == RestraintsFuncRMSD) then
          enefunc%rmsd_force  = enefunc%restraint_const(1,funcid)
          enefunc%rmsd_target = enefunc%restraint_ref  (1,funcid)
        end if
      end do

      ! compute restraint energy
      !
      call compute_energy_restraints_remd(.false., enefunc, domain, eneval)

    else
      eneval = 0.0_wp
    end if

    before_gather = eneval
#ifdef MPI
    call mpi_allgather(before_gather, 1, mpi_wp_real, &
                       after_gather,  1, mpi_wp_real, &
                       mpi_comm_airplane, ierror)
#endif
    do i = 1, total_nreplicas
      remd%restraint_energy_1(i) = after_gather(i)
    end do


    ! replica exchange
    !
    do i = 1, total_nreplicas

      repid_i     = i
      parmsetid_i = repid2parmsetid(repid_i)

      ! get neighboring parameter set index and replica index
      !
      call get_neighbouring_parmsetid(dimno, exptrn, remd, parmsetid_i, neighid)

      if (neighid /= 0) then
        parmsetid_j = neighid
        repid_j     = parmsetid2repid(parmsetid_j)
        do_trial    = .true.
      else
        do_trial    = .false.
      end if


      if (do_trial) then

        num_criteria(parmsetid_i,dimno,exptrn) &
          = num_criteria(parmsetid_i,dimno,exptrn) + 1

        ! compute delta
        !
        call get_replica_temperature(parmsetid_i, remd, ensemble, temperature)

        energy_i_0 = remd%restraint_energy_0(repid_i)
        energy_i_1 = remd%restraint_energy_1(repid_i)
        energy_j_0 = remd%restraint_energy_0(repid_j)
        energy_j_1 = remd%restraint_energy_1(repid_j)
        beta = 1.0_wp / (KBOLTZ * temperature)
        delta = beta*(energy_j_1 - energy_i_0 - energy_j_0 + energy_i_1)

        ! Metropolis criterion
        !
        if (delta <= 0.0_wp) then
          do_exchange = .true.
        else if (delta > 0.0_wp) then
          if (random_table(repid_i,repid_j) <= exp(-delta)) then
            do_exchange = .true.
          else
            do_exchange = .false.
          end if
        end if

        if (do_exchange) then
          num_exchanges(parmsetid_i,dimno,exptrn) &
            = num_exchanges(parmsetid_i,dimno,exptrn) + 1
          repid2parmsetid(i) = parmsetid_j
        end if

      end if
    end do


    return

  end subroutine restraint_remd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    assign_condition
  !> @brief        control replica exchange
  !! @authors      TM
  !! @param[inout] molecule    : molecule information
  !! @param[inout] enefunc     : potential energy functions information
  !! @param[inout] dynvars     : dynamics variables information
  !! @param[inout] boundary    : boundary conditions information
  !! @param[inout] ensemble    : ensemble information
  !! @param[inout] remd        : REMD information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine assign_condition(parmsetid, remd, domain, ensemble, enefunc)

    ! formal arguments
    integer,                 intent(in)    :: parmsetid
    type(s_remd),            intent(in)    :: remd
    type(s_domain),          intent(in)    :: domain
    type(s_ensemble),        intent(inout) :: ensemble
    type(s_enefunc),         intent(inout) :: enefunc

    ! local variables
    integer                  :: i, j, k
    integer                  :: umbrid, funcid, atmid


    do i = 1, remd%dimension
      if      (remd%types(i) == RemdTemperature) then
        ensemble%temperature = remd%dparameters(i,remd%parmidsets(parmsetid,i))
      else if (remd%types(i) == RemdPressure   ) then
        ensemble%pressure    = remd%dparameters(i,remd%parmidsets(parmsetid,i))
      else if (remd%types(i) == RemdGamma      ) then
        ensemble%gamma       = remd%dparameters(i,remd%parmidsets(parmsetid,i))
      else if (remd%types(i) == RemdRestraint  ) then
        umbrid = remd%iparameters(i,remd%parmidsets(parmsetid,i))
        do k = 1, remd%umbrid2numfuncs(umbrid)
          funcid = remd%umbrid2funclist(umbrid,k)
          enefunc%restraint_const(1:2,funcid) = remd%rest_constants(umbrid,k)
          enefunc%restraint_ref  (1:2,funcid) = remd%rest_reference(umbrid,k)

          if (enefunc%restraint_kind(funcid) == RestraintsFuncRMSD) then
            enefunc%rmsd_force  = enefunc%restraint_const(1,funcid)
            enefunc%rmsd_target = enefunc%restraint_ref  (1,funcid)
          end if

        end do
      end if
    end do

    return

  end subroutine assign_condition

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    get_neighbouring_parmsetid
  !> @brief        get neighbouring parameter set id
  !! @authors      TM
  !! @param[in]    dimno        : dimension number
  !! @param[in]    exptrn       : exchange pattern
  !! @param[in]    remd         : REMD information
  !! @param[in]    parmsetid_i  : parameter set id
  !! @param[out]   neighbour_id : parameter set id neighbouring parmsetid_i
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine get_neighbouring_parmsetid(dimno, exptrn, remd, parmsetid_i, &
                                        neighbour_id)

    ! formal arguments
    integer,                 intent(in)    :: dimno
    integer,                 intent(in)    :: exptrn
    type(s_remd),    target, intent(in)    :: remd
    integer,                 intent(in)    :: parmsetid_i
    integer,                 intent(out)   :: neighbour_id

    ! local variables
    integer                  :: j, k
    integer                  :: next, jump, parmid_i
    logical                  :: neighbour1, neighbour2
    logical                  :: cyclic
    integer,         pointer :: dimension, num_parmidsets
    integer,         pointer :: parmidsets(:,:)
    

    cyclic = remd%cyclic_params(dimno)

    dimension      => remd%dimension
    num_parmidsets => remd%total_nreplicas
    parmidsets     => remd%parmidsets

    parmid_i = parmidsets(parmsetid_i,dimno)
    if (exptrn == 1) then
      if (mod(parmid_i,2) == 0) then
        ! 1 <2  3 <4  5 <6  7 <8  ...
        next = -1
        if (cyclic) then
          jump = remd%nreplicas(dimno) - 1
        else
          jump = next
        end if
      else
        ! 1> 2  3> 4  5> 6  7> 8  ...
        next = +1
        if (cyclic) then
          jump = -remd%nreplicas(dimno) + 1
        else
          jump = next
        end if
      end if
    else if (exptrn == 2) then
      if (mod(parmid_i,2) == 0) then
        ! 1  2> 3  4> 5  6> 7  8  ...
        next = +1
        if (cyclic) then
          jump = -remd%nreplicas(dimno) + 1
        else
          jump = next
        end if
      else
        ! 1  2 <3  4 <5  6 <7  8  ...
        next = -1
        if (cyclic) then
          jump = remd%nreplicas(dimno) - 1
        else
          jump = next
        end if
      end if
    end if

    do j = 1, num_parmidsets
      neighbour1 = .true.
      neighbour2 = .false.
      do k = 1, dimension
        if (k /= dimno) then
          if (parmidsets(j,k) /= parmidsets(parmsetid_i,k)) &
            neighbour1 = .false.
        else
          if (parmidsets(j,k) == parmidsets(parmsetid_i,k) + next) then
            neighbour2 = .true.
          else if (parmidsets(j,k) == parmidsets(parmsetid_i,k) + jump) then
            neighbour2 = .true.
          end if
        end if
      end do
      if (neighbour1 .and. neighbour2) then
        neighbour_id = j
        exit
      else
        neighbour_id = 0
      end if
    end do

    return

  end subroutine get_neighbouring_parmsetid

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    get_replica_temperature
  !> @brief        get temperature of each replica
  !! @authors      TM
  !! @param[in]    parmsetid   : parameter set id
  !! @param[in]    remd        : REMD information
  !! @param[in]    ensemble    : ensemble information
  !! @param[out]   temperature : temperature of each replica
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine get_replica_temperature(parmsetid, remd, ensemble, temperature)

    ! formal arguments
    integer,          intent(in)  :: parmsetid
    type(s_remd),     intent(in)  :: remd
    type(s_ensemble), intent(in)  :: ensemble
    real(wp),         intent(out) :: temperature

    ! local variables
    integer                       :: i

    temperature = ensemble%temperature

    do i = 1, remd%dimension
      if (remd%types(i) == RemdTemperature) then
        temperature = remd%dparameters(i,remd%parmidsets(parmsetid,i))
      end if
    end do

    return

  end subroutine get_replica_temperature

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_restraints_remd
  !> @brief        compute restraint potential energy in REUS routines
  !! @authors      TM
  !! @param[in]    domain : domain information
  !! @param[in]    enefuc : enefunc information
  !! @param[out]   eneval : energy value
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_restraints_remd(get_coord, enefunc, domain, eneval)

    ! formal arguments
    logical,                  intent(in)    :: get_coord
    type(s_enefunc),  target, intent(inout) :: enefunc
    type(s_domain),   target, intent(inout) :: domain
    real(wp),                 intent(out)   :: eneval

    ! local variables
    integer                  :: i, id
    logical                  :: do_posres
    real(dp)                 :: ebonds, eposi, eposi_mpi, ermsd
    real(dp)                 :: rmsd
    real(dp)                 :: eposi_omp(nthread)
    real(dp)                 :: virial_omp(3,3,nthread)
    real(dp)                 :: virial_ext_omp(3,3,nthread)
    real(wp),        pointer :: force_omp(:,:,:,:)


    force_omp  => domain%force_omp

    eneval     = 0.0_wp
    ebonds     = 0.0_wp
    ermsd      = 0.0_wp
    rmsd       = 0.0_dp

    eposi_omp(1:nthread) = 0.0_wp

    call compute_energy_restraints(get_coord, .false., domain, enefunc, &
                                   domain%coord, force_omp, virial_omp, &
                                   virial_ext_omp, eposi_omp,           &
                                   ermsd, rmsd, ebonds)

    eneval = ebonds + ermsd

    return

  end subroutine compute_energy_restraints_remd

end module sp_remd_mod
